@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -41,6 +41,7 @@
 #include "gatppp.h"
 
 #include "atmodem.h"
+#include "vendor.h"
 
 #define TUN_SYSFS_DIR "/sys/devices/virtual/misc/tun"
 
@@ -64,6 +65,7 @@ struct gprs_context_data {
 	enum state state;
 	ofono_gprs_context_cb_t cb;
 	void *cb_data;                                  /* Callback data */
+	unsigned int vendor;
 };
 
 static void ppp_debug(const char *str, void *data)
@@ -239,6 +241,27 @@ static void at_gprs_activate_primary(struct ofono_gprs_context *gc,
 
 	gcd->state = STATE_ENABLING;
 
+	if (gcd->vendor == OFONO_VENDOR_ZTE) {
+		GAtChat *chat = g_at_chat_get_slave(gcd->chat);
+
+		/*
+		 * The modem port of ZTE devices with certain firmware
+		 * versions ends up getting suspended. It will no longer
+		 * signal POLLOUT and becomes pretty unresponsive.
+		 *
+		 * To wake up the modem port, the only reliable method
+		 * found so far is AT+ZOPRT power mode command. It is
+		 * enough to ask for the current mode and the modem
+		 * port wakes up and accepts commands again.
+		 *
+		 * And since the modem port is suspended, this command
+		 * needs to be send on the control port of course.
+		 *
+		 */
+		g_at_chat_send(chat, "AT+ZOPRT?", none_prefix,
+						NULL, NULL, NULL);
+	}
+
 	len = snprintf(buf, sizeof(buf), "AT+CGDCONT=%u,\"IP\"", ctx->cid);
 
 	if (ctx->apn)
@@ -268,6 +291,50 @@ static void at_gprs_deactivate_primary(struct ofono_gprs_context *gc,
 	g_at_ppp_shutdown(gcd->ppp);
 }
 
+static void at_gprs_detach_shutdown(struct ofono_gprs_context *gc,
+					unsigned int cid)
+{
+	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
+
+	DBG("cid %u", cid);
+
+	g_at_ppp_shutdown(gcd->ppp);
+}
+
+static void cgev_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_gprs_context *gc = user_data;
+	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
+	const char *event;
+	int cid;
+	GAtResultIter iter;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "+CGEV:"))
+		return;
+
+	if (!g_at_result_iter_next_unquoted_string(&iter, &event))
+		return;
+
+	if (g_str_has_prefix(event, "NW DEACT") == FALSE)
+		return;
+
+	if (!g_at_result_iter_skip_next(&iter))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &cid))
+		return;
+
+	DBG("cid %d", cid);
+
+	if ((unsigned int) cid != gcd->active_context)
+		return;
+
+	if (gcd->state != STATE_IDLE && gcd->ppp)
+		g_at_ppp_shutdown(gcd->ppp);
+}
+
 static int at_gprs_context_probe(struct ofono_gprs_context *gc,
 					unsigned int vendor, void *data)
 {
@@ -287,9 +354,16 @@ static int at_gprs_context_probe(struct ofono_gprs_context *gc,
 		return -ENOMEM;
 
 	gcd->chat = g_at_chat_clone(chat);
+	gcd->vendor = vendor;
 
 	ofono_gprs_context_set_data(gc, gcd);
 
+	chat = g_at_chat_get_slave(gcd->chat);
+	if (chat == NULL)
+		return 0;
+
+	g_at_chat_register(chat, "+CGEV:", cgev_notify, FALSE, gc, NULL);
+
 	return 0;
 }
 
@@ -316,6 +390,7 @@ static struct ofono_gprs_context_driver driver = {
 	.remove			= at_gprs_context_remove,
 	.activate_primary	= at_gprs_activate_primary,
 	.deactivate_primary	= at_gprs_deactivate_primary,
+	.detach_shutdown	= at_gprs_detach_shutdown,
 };
 
 void at_gprs_context_init(void)
