@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -34,37 +34,36 @@
 #include "ofono.h"
 
 #include "common.h"
+#include "smsutil.h"
+#include "util.h"
 
-#define SUPPLEMENTARY_SERVICES_INTERFACE "org.ofono.SupplementaryServices"
-
-#define USSD_FLAG_PENDING 0x1
+#define MAX_USSD_LENGTH 160
 
 static GSList *g_drivers = NULL;
 
-enum ussd_status {
-	USSD_STATUS_NOTIFY = 0,
-	USSD_STATUS_ACTION_REQUIRED = 1,
-	USSD_STATUS_TERMINATED = 2,
-	USSD_STATUS_LOCAL_CLIENT_RESPONDED = 3,
-	USSD_STATUS_NOT_SUPPORTED = 4,
-	USSD_STATUS_TIMED_OUT = 5,
-};
-
 enum ussd_state {
 	USSD_STATE_IDLE = 0,
 	USSD_STATE_ACTIVE = 1,
-	USSD_STATE_USER_ACTION = 2
+	USSD_STATE_USER_ACTION = 2,
+	USSD_STATE_RESPONSE_SENT,
+};
+
+struct ussd_request {
+	ofono_ussd_request_cb_t cb;
+	void *user_data;
 };
 
 struct ofono_ussd {
 	int state;
 	DBusMessage *pending;
+	DBusMessage *cancel;
 	int flags;
 	GSList *ss_control_list;
 	GSList *ss_passwd_list;
 	const struct ofono_ussd_driver *driver;
 	void *driver_data;
 	struct ofono_atom *atom;
+	struct ussd_request *req;
 };
 
 struct ssc_entry {
@@ -74,6 +73,17 @@ struct ssc_entry {
 	ofono_destroy_func destroy;
 };
 
+gboolean __ofono_ussd_is_busy(struct ofono_ussd *ussd)
+{
+	if (ussd == NULL)
+		return FALSE;
+
+	if (ussd->pending || ussd->state != USSD_STATE_IDLE || ussd->req)
+		return TRUE;
+
+	return FALSE;
+}
+
 static struct ssc_entry *ssc_entry_create(const char *sc, void *cb, void *data,
 						ofono_destroy_func destroy)
 {
@@ -81,7 +91,7 @@ static struct ssc_entry *ssc_entry_create(const char *sc, void *cb, void *data,
 
 	r = g_try_new0(struct ssc_entry, 1);
 
-	if (!r)
+	if (r == NULL)
 		return r;
 
 	r->service = g_strdup(sc);
@@ -114,12 +124,11 @@ gboolean __ofono_ussd_ssc_register(struct ofono_ussd *ussd, const char *sc,
 {
 	struct ssc_entry *entry;
 
-	if (!ussd)
+	if (ussd == NULL)
 		return FALSE;
 
 	entry = ssc_entry_create(sc, cb, data, destroy);
-
-	if (!entry)
+	if (entry == NULL)
 		return FALSE;
 
 	ussd->ss_control_list = g_slist_prepend(ussd->ss_control_list, entry);
@@ -131,13 +140,13 @@ void __ofono_ussd_ssc_unregister(struct ofono_ussd *ussd, const char *sc)
 {
 	GSList *l;
 
-	if (!ussd)
+	if (ussd == NULL)
 		return;
 
 	l = g_slist_find_custom(ussd->ss_control_list, sc,
 				ssc_entry_find_by_service);
 
-	if (!l)
+	if (l == NULL)
 		return;
 
 	ssc_entry_destroy(l->data);
@@ -150,12 +159,11 @@ gboolean __ofono_ussd_passwd_register(struct ofono_ussd *ussd, const char *sc,
 {
 	struct ssc_entry *entry;
 
-	if (!ussd)
+	if (ussd == NULL)
 		return FALSE;
 
 	entry = ssc_entry_create(sc, cb, data, destroy);
-
-	if (!entry)
+	if (entry == NULL)
 		return FALSE;
 
 	ussd->ss_passwd_list = g_slist_prepend(ussd->ss_passwd_list, entry);
@@ -167,13 +175,13 @@ void __ofono_ussd_passwd_unregister(struct ofono_ussd *ussd, const char *sc)
 {
 	GSList *l;
 
-	if (!ussd)
+	if (ussd == NULL)
 		return;
 
 	l = g_slist_find_custom(ussd->ss_passwd_list, sc,
 				ssc_entry_find_by_service);
 
-	if (!l)
+	if (l == NULL)
 		return;
 
 	ssc_entry_destroy(l->data);
@@ -231,18 +239,20 @@ static gboolean recognized_control_string(struct ofono_ussd *ussd,
 	int type;
 	gboolean ret = FALSE;
 
-	ofono_debug("parsing control string");
+	DBG("parsing control string");
 
 	if (parse_ss_control_string(str, &type, &sc,
 				&sia, &sib, &sic, &sid, &dn)) {
 		GSList *l = ussd->ss_control_list;
 
-		ofono_debug("Got parse result: %d, %s, %s, %s, %s, %s, %s",
+		DBG("Got parse result: %d, %s, %s, %s, %s, %s, %s",
 				type, sc, sia, sib, sic, sid, dn);
 
-		/* A password change string needs to be treated separately
+		/*
+		 * A password change string needs to be treated separately
 		 * because it uses a fourth SI and is thus not a valid
-		 * control string.  */
+		 * control string.
+		 */
 		if (recognized_passwd_change_string(ussd, type, sc,
 					sia, sib, sic, sid, dn, msg)) {
 			ret = TRUE;
@@ -269,10 +279,10 @@ static gboolean recognized_control_string(struct ofono_ussd *ussd,
 
 	/* TODO: Handle all strings that control voice calls */
 
-	/* TODO: Handle Multiple subscriber profile DN*59#SEND and *59#SEND
-	 */
+	/* TODO: Handle Multiple subscriber profile DN*59#SEND and *59#SEND */
 
-	/* Note: SIM PIN/PIN2 change and unblock and IMEI presentation
+	/*
+	 * Note: SIM PIN/PIN2 change and unblock and IMEI presentation
 	 * procedures are not handled by the daemon since they are not followed
 	 * by SEND and are not valid USSD requests.
 	 */
@@ -283,37 +293,158 @@ out:
 	return ret;
 }
 
-void ofono_ussd_notify(struct ofono_ussd *ussd, int status, const char *str)
+static const char *ussd_get_state_string(struct ofono_ussd *ussd)
+{
+	switch (ussd->state) {
+	case USSD_STATE_IDLE:
+		return "idle";
+	case USSD_STATE_ACTIVE:
+	case USSD_STATE_RESPONSE_SENT:
+		return "active";
+	case USSD_STATE_USER_ACTION:
+		return "user-response";
+	}
+
+	return "";
+}
+
+static void ussd_change_state(struct ofono_ussd *ussd, int state)
+{
+	const char *value;
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(ussd->atom);
+
+	if (state == ussd->state)
+		return;
+
+	ussd->state = state;
+
+	value = ussd_get_state_string(ussd);
+	ofono_dbus_signal_property_changed(conn, path,
+			OFONO_SUPPLEMENTARY_SERVICES_INTERFACE,
+			"State", DBUS_TYPE_STRING, &value);
+}
+
+static void ussd_request_finish(struct ofono_ussd *ussd, int error, int dcs,
+				const unsigned char *pdu, int len)
+{
+	struct ussd_request *req = ussd->req;
+
+	if (req && req->cb)
+		req->cb(error, dcs, pdu, len, req->user_data);
+
+	g_free(req);
+	ussd->req = NULL;
+}
+
+static int ussd_status_to_failure_code(int status)
+{
+	switch (status) {
+	case OFONO_USSD_STATUS_TIMED_OUT:
+		return -ETIMEDOUT;
+	case OFONO_USSD_STATUS_NOT_SUPPORTED:
+		return -ENOSYS;
+	}
+
+	return 0;
+}
+
+static char const *ussd_status_name(int status)
+{
+	switch (status) {
+	case OFONO_USSD_STATUS_NOTIFY:
+		return "NOTIFY";
+	case OFONO_USSD_STATUS_ACTION_REQUIRED:
+		return "ACTION_REQUIRED";
+	case OFONO_USSD_STATUS_TERMINATED:
+		return "TERMINATED";
+	case OFONO_USSD_STATUS_LOCAL_CLIENT_RESPONDED:
+		return "LOCAL_CLIENT_RESPONDED";
+	case OFONO_USSD_STATUS_NOT_SUPPORTED:
+		return "NOT_SUPPORTED";
+	case OFONO_USSD_STATUS_TIMED_OUT:
+		return "TIMED_OUT";
+	}
+
+	return "????";
+}
+
+static const char *ussd_state_name(enum ussd_state state)
+{
+	switch (state) {
+	case USSD_STATE_IDLE:
+		return "IDLE";
+	case USSD_STATE_ACTIVE:
+		return "ACTIVE";
+	case USSD_STATE_RESPONSE_SENT:
+		return "RESPONSE_SENT";
+	case USSD_STATE_USER_ACTION:
+		return "USER_ACTION";
+	}
+
+	return "????";
+}
+
+
+void ofono_ussd_notify(struct ofono_ussd *ussd, int status, int dcs,
+			const unsigned char *data, int data_len)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *ussdstr = "USSD";
+	char *utf8_str = NULL;
+	const char *str;
 	const char sig[] = { DBUS_TYPE_STRING, 0 };
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter variant;
 
-	if (status == USSD_STATUS_NOT_SUPPORTED) {
-		ussd->state = USSD_STATE_IDLE;
+	DBG("status: %d %s, state: %d %s",
+		status, ussd_status_name(status),
+		ussd->state, ussd_state_name(ussd->state));
+
+	if (ussd->req &&
+			(status == OFONO_USSD_STATUS_NOTIFY ||
+			status == OFONO_USSD_STATUS_TERMINATED ||
+			status == OFONO_USSD_STATUS_TIMED_OUT ||
+			status == OFONO_USSD_STATUS_NOT_SUPPORTED)) {
+		ussd_request_finish(ussd, ussd_status_to_failure_code(status),
+					dcs, data, data_len);
+
+		ussd_change_state(ussd, USSD_STATE_IDLE);
+		return;
+	}
+
+	if (status == OFONO_USSD_STATUS_NOT_SUPPORTED) {
+		ussd_change_state(ussd, USSD_STATE_IDLE);
+
+		if (ussd->pending == NULL)
+			return;
+
 		reply = __ofono_error_not_supported(ussd->pending);
 		goto out;
 	}
 
-	if (status == USSD_STATUS_TIMED_OUT) {
-		ussd->state = USSD_STATE_IDLE;
+	if (status == OFONO_USSD_STATUS_TIMED_OUT) {
+		ussd_change_state(ussd, USSD_STATE_IDLE);
+
+		if (ussd->pending == NULL)
+			return;
+
 		reply = __ofono_error_timed_out(ussd->pending);
 		goto out;
 	}
 
+	if (data && data_len > 0)
+		utf8_str = ussd_decode(dcs, data_len, data);
+
+	str = utf8_str;
+
 	/* TODO: Rework this in the Agent framework */
 	if (ussd->state == USSD_STATE_ACTIVE) {
-		if (status == USSD_STATUS_ACTION_REQUIRED) {
-			ofono_error("Unable to handle action required ussd");
-			return;
-		}
 
 		reply = dbus_message_new_method_return(ussd->pending);
 
-		if (!str)
+		if (str == NULL)
 			str = "";
 
 		dbus_message_iter_init_append(reply, &iter);
@@ -329,12 +460,51 @@ void ofono_ussd_notify(struct ofono_ussd *ussd, int status, const char *str)
 
 		dbus_message_iter_close_container(&iter, &variant);
 
-		ussd->state = USSD_STATE_IDLE;
-	} else {
-		ofono_error("Received an unsolicited USSD, ignoring for now...");
-		ofono_debug("USSD is: status: %d, %s", status, str);
+		if (status == OFONO_USSD_STATUS_ACTION_REQUIRED)
+			ussd_change_state(ussd, USSD_STATE_USER_ACTION);
+		else
+			ussd_change_state(ussd, USSD_STATE_IDLE);
 
-		return;
+	} else if (ussd->state == USSD_STATE_RESPONSE_SENT) {
+		reply = dbus_message_new_method_return(ussd->pending);
+
+		if (str == NULL)
+			str = "";
+
+		dbus_message_append_args(reply, DBUS_TYPE_STRING, &str,
+						DBUS_TYPE_INVALID);
+
+		if (status == OFONO_USSD_STATUS_ACTION_REQUIRED)
+			ussd_change_state(ussd, USSD_STATE_USER_ACTION);
+		else
+			ussd_change_state(ussd, USSD_STATE_IDLE);
+	} else if (ussd->state == USSD_STATE_IDLE) {
+		const char *signal_name;
+		const char *path = __ofono_atom_get_path(ussd->atom);
+		int new_state;
+
+		if (status == OFONO_USSD_STATUS_ACTION_REQUIRED) {
+			new_state = USSD_STATE_USER_ACTION;
+			signal_name = "RequestReceived";
+		} else {
+			new_state = USSD_STATE_IDLE;
+			signal_name = "NotificationReceived";
+		}
+
+		if (str == NULL)
+			str = "";
+
+		g_dbus_emit_signal(conn, path,
+			OFONO_SUPPLEMENTARY_SERVICES_INTERFACE, signal_name,
+				DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID);
+
+		ussd_change_state(ussd, new_state);
+		goto free;
+	} else {
+		ofono_error("Received an unsolicited USSD but can't handle.");
+		DBG("USSD is: status: %d, %s", status, str);
+
+		goto free;
 	}
 
 out:
@@ -342,46 +512,45 @@ out:
 
 	dbus_message_unref(ussd->pending);
 	ussd->pending = NULL;
+
+free:
+	g_free(utf8_str);
 }
 
 static void ussd_callback(const struct ofono_error *error, void *data)
 {
 	struct ofono_ussd *ussd = data;
-	DBusConnection *conn = ofono_dbus_get_connection();
 	DBusMessage *reply;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
-		ofono_debug("ussd request failed with error: %s",
+		DBG("ussd request failed with error: %s",
 				telephony_error_to_str(error));
 
-	ussd->flags &= ~USSD_FLAG_PENDING;
-
-	if (!ussd->pending)
-		return;
-
 	if (error->type == OFONO_ERROR_TYPE_NO_ERROR) {
-		ussd->state = USSD_STATE_ACTIVE;
+		ussd_change_state(ussd, USSD_STATE_ACTIVE);
 		return;
 	}
 
+	if (ussd->pending == NULL)
+		return;
+
 	reply = __ofono_error_failed(ussd->pending);
-
-	g_dbus_send_message(conn, reply);
-
-	dbus_message_unref(ussd->pending);
-	ussd->pending = NULL;
+	__ofono_dbus_pending_reply(&ussd->pending, reply);
 }
 
 static DBusMessage *ussd_initiate(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
 	struct ofono_ussd *ussd = data;
+	struct ofono_modem *modem = __ofono_atom_get_modem(ussd->atom);
+	struct ofono_voicecall *vc;
+	gboolean call_in_progress;
 	const char *str;
+	int dcs = 0x0f;
+	unsigned char buf[160];
+	long num_packed;
 
-	if (ussd->flags & USSD_FLAG_PENDING)
-		return __ofono_error_busy(msg);
-
-	if (ussd->state == USSD_STATE_ACTIVE)
+	if (__ofono_ussd_is_busy(ussd))
 		return __ofono_error_busy(msg);
 
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &str,
@@ -391,23 +560,89 @@ static DBusMessage *ussd_initiate(DBusConnection *conn, DBusMessage *msg,
 	if (strlen(str) == 0)
 		return __ofono_error_invalid_format(msg);
 
-	ofono_debug("checking if this is a recognized control string");
+	DBG("checking if this is a recognized control string");
 	if (recognized_control_string(ussd, str, msg))
 		return NULL;
 
-	ofono_debug("No.., checking if this is a USSD string");
-	if (!valid_ussd_string(str))
+	vc = __ofono_atom_find(OFONO_ATOM_TYPE_VOICECALL, modem);
+	if (vc)
+		call_in_progress = __ofono_voicecall_is_busy(vc,
+					OFONO_VOICECALL_INTERACTION_NONE);
+	else
+		call_in_progress = FALSE;
+
+	DBG("No.., checking if this is a USSD string");
+	if (!valid_ussd_string(str, call_in_progress))
+		return __ofono_error_not_recognized(msg);
+
+	if (!ussd_encode(str, &num_packed, buf))
 		return __ofono_error_invalid_format(msg);
 
-	ofono_debug("OK, running USSD request");
-
-	if (!ussd->driver->request)
+	if (ussd->driver->request == NULL)
 		return __ofono_error_not_implemented(msg);
 
-	ussd->flags |= USSD_FLAG_PENDING;
+	DBG("OK, running USSD request");
+
 	ussd->pending = dbus_message_ref(msg);
 
-	ussd->driver->request(ussd, str, ussd_callback, ussd);
+	ussd->driver->request(ussd, dcs, buf, num_packed, ussd_callback, ussd);
+
+	return NULL;
+}
+
+static void ussd_response_callback(const struct ofono_error *error, void *data)
+{
+	struct ofono_ussd *ussd = data;
+	DBusMessage *reply;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		DBG("ussd response failed with error: %s",
+				telephony_error_to_str(error));
+
+	if (error->type == OFONO_ERROR_TYPE_NO_ERROR) {
+		ussd_change_state(ussd, USSD_STATE_RESPONSE_SENT);
+		return;
+	}
+
+	if (ussd->pending == NULL)
+		return;
+
+	reply = __ofono_error_failed(ussd->pending);
+	__ofono_dbus_pending_reply(&ussd->pending, reply);
+}
+
+static DBusMessage *ussd_respond(DBusConnection *conn, DBusMessage *msg,
+					void *data)
+{
+	struct ofono_ussd *ussd = data;
+	const char *str;
+	int dcs = 0x0f;
+	unsigned char buf[160];
+	long num_packed;
+
+	if (ussd->pending)
+		return __ofono_error_busy(msg);
+
+	if (ussd->state != USSD_STATE_USER_ACTION)
+		return __ofono_error_not_active(msg);
+
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &str,
+					DBUS_TYPE_INVALID) == FALSE)
+		return __ofono_error_invalid_args(msg);
+
+	if (strlen(str) == 0)
+		return __ofono_error_invalid_format(msg);
+
+	if (!ussd_encode(str, &num_packed, buf))
+		return __ofono_error_invalid_format(msg);
+
+	if (ussd->driver->request == NULL)
+		return __ofono_error_not_implemented(msg);
+
+	ussd->pending = dbus_message_ref(msg);
+
+	ussd->driver->request(ussd, dcs, buf, num_packed,
+				ussd_response_callback, ussd);
 
 	return NULL;
 }
@@ -417,24 +652,28 @@ static void ussd_cancel_callback(const struct ofono_error *error, void *data)
 	struct ofono_ussd *ussd = data;
 	DBusMessage *reply;
 
-	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
-		ofono_debug("ussd cancel failed with error: %s",
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		DBG("ussd cancel failed with error: %s",
 				telephony_error_to_str(error));
 
-	ussd->flags &= ~USSD_FLAG_PENDING;
+		reply = __ofono_error_failed(ussd->cancel);
+		__ofono_dbus_pending_reply(&ussd->cancel, reply);
 
-	if (!ussd->pending)
 		return;
-
-	if (error->type == OFONO_ERROR_TYPE_NO_ERROR) {
-		ussd->state = USSD_STATE_IDLE;
-
-		reply = dbus_message_new_method_return(ussd->pending);
-	} else {
-		reply = __ofono_error_failed(ussd->pending);
 	}
 
-	__ofono_dbus_pending_reply(&ussd->pending, reply);
+	if (ussd->pending) {
+		reply = __ofono_error_canceled(ussd->pending);
+		__ofono_dbus_pending_reply(&ussd->pending, reply);
+	}
+
+	reply = dbus_message_new_method_return(ussd->cancel);
+	__ofono_dbus_pending_reply(&ussd->cancel, reply);
+
+	if (ussd->req)
+		ussd_request_finish(ussd, -ECANCELED, 0, NULL, 0);
+
+	ussd_change_state(ussd, USSD_STATE_IDLE);
 }
 
 static DBusMessage *ussd_cancel(DBusConnection *conn, DBusMessage *msg,
@@ -442,32 +681,76 @@ static DBusMessage *ussd_cancel(DBusConnection *conn, DBusMessage *msg,
 {
 	struct ofono_ussd *ussd = data;
 
-	if (ussd->flags & USSD_FLAG_PENDING)
-		return __ofono_error_busy(msg);
-
 	if (ussd->state == USSD_STATE_IDLE)
 		return __ofono_error_not_active(msg);
 
-	if (!ussd->driver->cancel)
+	/* We have called Respond() but not returned from its callback yet */
+	if (ussd->state == USSD_STATE_USER_ACTION && ussd->pending)
+		return __ofono_error_busy(msg);
+
+	if (ussd->cancel)
+		return __ofono_error_busy(msg);
+
+	if (ussd->driver->cancel == NULL)
 		return __ofono_error_not_implemented(msg);
 
-	ussd->flags |= USSD_FLAG_PENDING;
-	ussd->pending = dbus_message_ref(msg);
+	ussd->cancel = dbus_message_ref(msg);
 
 	ussd->driver->cancel(ussd, ussd_cancel_callback, ussd);
 
 	return NULL;
 }
 
-static GDBusMethodTable ussd_methods[] = {
-	{ "Initiate",	"s",	"sv",	ussd_initiate,
-					G_DBUS_METHOD_FLAG_ASYNC },
-	{ "Cancel",	"",	"",	ussd_cancel,
-					G_DBUS_METHOD_FLAG_ASYNC },
+static DBusMessage *ussd_get_properties(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	struct ofono_ussd *ussd = data;
+	DBusMessage *reply;
+	DBusMessageIter iter;
+	DBusMessageIter dict;
+	const char *value;
+
+	reply = dbus_message_new_method_return(msg);
+	if (reply == NULL)
+		return NULL;
+
+	dbus_message_iter_init_append(reply, &iter);
+
+	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
+					OFONO_PROPERTIES_ARRAY_SIGNATURE,
+					&dict);
+
+	value = ussd_get_state_string(ussd);
+	ofono_dbus_dict_append(&dict, "State", DBUS_TYPE_STRING, &value);
+
+	dbus_message_iter_close_container(&iter, &dict);
+
+	return reply;
+}
+
+static const GDBusMethodTable ussd_methods[] = {
+	{ GDBUS_ASYNC_METHOD("Initiate",
+			GDBUS_ARGS({ "command", "s" }),
+			GDBUS_ARGS({ "result_name", "s" }, { "value", "v" }),
+			ussd_initiate) },
+	{ GDBUS_ASYNC_METHOD("Respond",
+			GDBUS_ARGS({ "reply", "s" }),
+			GDBUS_ARGS({ "result", "s" }),
+			ussd_respond) },
+	{ GDBUS_ASYNC_METHOD("Cancel", NULL, NULL, ussd_cancel) },
+	{ GDBUS_METHOD("GetProperties",
+			NULL, GDBUS_ARGS({ "properties", "a{sv}" }),
+			ussd_get_properties) },
 	{ }
 };
 
-static GDBusSignalTable ussd_signals[] = {
+static const GDBusSignalTable ussd_signals[] = {
+	{ GDBUS_SIGNAL("NotificationReceived",
+					GDBUS_ARGS({ "message", "s" })) },
+	{ GDBUS_SIGNAL("RequestReceived",
+					GDBUS_ARGS({ "message", "s" })) },
+	{ GDBUS_SIGNAL("PropertyChanged",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" })) },
 	{ }
 };
 
@@ -478,7 +761,7 @@ int ofono_ussd_driver_register(const struct ofono_ussd_driver *d)
 	if (d->probe == NULL)
 		return -EINVAL;
 
-	g_drivers = g_slist_prepend(g_drivers, (void *)d);
+	g_drivers = g_slist_prepend(g_drivers, (void *) d);
 
 	return 0;
 }
@@ -487,7 +770,7 @@ void ofono_ussd_driver_unregister(const struct ofono_ussd_driver *d)
 {
 	DBG("driver: %p, name: %s", d, d->name);
 
-	g_drivers = g_slist_remove(g_drivers, (void *)d);
+	g_drivers = g_slist_remove(g_drivers, (void *) d);
 }
 
 static void ussd_unregister(struct ofono_atom *atom)
@@ -497,17 +780,18 @@ static void ussd_unregister(struct ofono_atom *atom)
 	struct ofono_modem *modem = __ofono_atom_get_modem(atom);
 	const char *path = __ofono_atom_get_path(atom);
 
-	g_slist_foreach(ussd->ss_control_list, (GFunc)ssc_entry_destroy, NULL);
+	g_slist_foreach(ussd->ss_control_list, (GFunc) ssc_entry_destroy, NULL);
 	g_slist_free(ussd->ss_control_list);
 	ussd->ss_control_list = NULL;
 
-	g_slist_foreach(ussd->ss_passwd_list, (GFunc)ssc_entry_destroy, NULL);
+	g_slist_foreach(ussd->ss_passwd_list, (GFunc) ssc_entry_destroy, NULL);
 	g_slist_free(ussd->ss_passwd_list);
 	ussd->ss_passwd_list = NULL;
 
-	ofono_modem_remove_interface(modem, SUPPLEMENTARY_SERVICES_INTERFACE);
+	ofono_modem_remove_interface(modem,
+					OFONO_SUPPLEMENTARY_SERVICES_INTERFACE);
 	g_dbus_unregister_interface(conn, path,
-					SUPPLEMENTARY_SERVICES_INTERFACE);
+					OFONO_SUPPLEMENTARY_SERVICES_INTERFACE);
 }
 
 static void ussd_remove(struct ofono_atom *atom)
@@ -567,16 +851,17 @@ void ofono_ussd_register(struct ofono_ussd *ussd)
 	const char *path = __ofono_atom_get_path(ussd->atom);
 
 	if (!g_dbus_register_interface(conn, path,
-					SUPPLEMENTARY_SERVICES_INTERFACE,
+					OFONO_SUPPLEMENTARY_SERVICES_INTERFACE,
 					ussd_methods, ussd_signals, NULL,
 					ussd, NULL)) {
 		ofono_error("Could not create %s interface",
-				SUPPLEMENTARY_SERVICES_INTERFACE);
+				OFONO_SUPPLEMENTARY_SERVICES_INTERFACE);
 
 		return;
 	}
 
-	ofono_modem_add_interface(modem, SUPPLEMENTARY_SERVICES_INTERFACE);
+	ofono_modem_add_interface(modem,
+				OFONO_SUPPLEMENTARY_SERVICES_INTERFACE);
 
 	__ofono_atom_register(ussd->atom, ussd_unregister);
 }
@@ -595,3 +880,47 @@ void *ofono_ussd_get_data(struct ofono_ussd *ussd)
 {
 	return ussd->driver_data;
 }
+
+static void ussd_request_callback(const struct ofono_error *error, void *data)
+{
+	struct ofono_ussd *ussd = data;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		ussd_request_finish(ussd, -EINVAL, 0, NULL, 0);
+	else
+		ussd_change_state(ussd, USSD_STATE_ACTIVE);
+}
+
+int __ofono_ussd_initiate(struct ofono_ussd *ussd, int dcs,
+				const unsigned char *pdu, int len,
+				ofono_ussd_request_cb_t cb, void *user_data)
+{
+	struct ussd_request *req;
+
+	if (ussd->driver->request == NULL)
+		return -ENOSYS;
+
+	if (__ofono_ussd_is_busy(ussd))
+		return -EBUSY;
+
+	req = g_try_new0(struct ussd_request, 1);
+	if (req == NULL)
+		return -ENOMEM;
+
+	req->cb = cb;
+	req->user_data = user_data;
+
+	ussd->req = req;
+
+	ussd->driver->request(ussd, dcs, pdu, len, ussd_request_callback, ussd);
+
+	return 0;
+}
+
+void __ofono_ussd_initiate_cancel(struct ofono_ussd *ussd)
+{
+	if (ussd->req == NULL || ussd->req->cb == NULL)
+		return;
+
+	ussd->req->cb = NULL;
+}
